/*
 * Copyright (C) 2012 Jens Georg <mail@jensge.org>.
 * Copyright (C) 2012, 2013 Intel Corporation.
 *
 * Author: Jens Georg <mail@jensge.org>
 *
 * This file is part of Rygel.
 *
 * Rygel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Rygel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "rygel-media-export-collate.h"

#ifdef HAVE_UNISTRING
#include <unistr.h>
gint
rygel_media_export_utf8_collate_str (const char *a,
                                     const char *b)
{
  return u8_strcoll ((const uint8_t *) a, (const uint8_t *) b);
}

#else

gint
rygel_media_export_utf8_collate_str (const char *a,
                                     const char *b)
{
  return g_utf8_collate (a, b);
}
#endif
